#include		"Intersect.h"

void clIntersect::funcGetIntervalSphere( clSphere a, clVec2 axis, float & fMin, float & fMax )
{
	fMin = ( a.vPos - axis * a.fRadius ) & axis;
	fMax = ( a.vPos + axis * a.fRadius ) & axis;

	if( fMin > fMax )
	{
		float	fTemp = fMin;
		fMin = fMax;
		fMax = fTemp;
	}
}

void clIntersect::funcGetIntervalLine( clLine a, clVec2 axis, float & fMin, float & fMax )
{
	clVec2	v;

	v = a.vMid - a.vDir * a.fHalf;
	fMin = v & axis;
	v = a.vMid + a.vDir * a.fHalf;
	fMax = v & axis;

	if( fMax < fMin )
	{
		float	fTemp = fMax;
		fMax = fMin;
		fMin = fTemp;
	}
}

void clIntersect::funcGetIntervalAABB( clAABB a, clVec2 axis, float & fMin, float & fMax )
{
	fMin = ( a.vPos - a.vSiz ) & ( axis );
	fMax = ( a.vPos - a.vSiz ) & ( axis );

	float	fTemp = clVec2( a.vPos.x - a.vSiz.x, a.vPos.y + a.vSiz.y ) & ( axis );
	if( fMin > fTemp )
	{
		fMin = fTemp;
	}
	else
	{
		if( fMax < fTemp )
		{
			fMax = fTemp;
		}
	}
	fTemp = ( a.vPos + a.vSiz ) & ( axis );
	if( fMin > fTemp )
	{
		fMin = fTemp;
	}
	else
	{
		if( fMax < fTemp )
		{
			fMax = fTemp;
		}
	}
	fTemp = clVec2( a.vPos.x + a.vSiz.x, a.vPos.y - a.vSiz.y ) & ( axis );
	if( fMin > fTemp )
	{
		fMin = fTemp;
	}
	else
	{
		if( fMax < fTemp )
		{
			fMax = fTemp;
		}
	}
}

void clIntersect::funcGetIntervalOBB( clOBB a, clVec2 axis, float & fMin, float & fMax )
{
	clVec2	v[ 4 ];
	clVec2	r[ 4 ];

	r[ 0 ].VecSet( - a.vSiz.x, - a.vSiz.y );
	r[ 1 ].VecSet( - a.vSiz.x, + a.vSiz.y );
	r[ 2 ].VecSet( + a.vSiz.x, + a.vSiz.y );
	r[ 3 ].VecSet( + a.vSiz.x, - a.vSiz.y );

	r[ 0 ].VecRotate( a.fAngle );
	r[ 1 ].VecRotate( a.fAngle );
	r[ 2 ].VecRotate( a.fAngle );
	r[ 3 ].VecRotate( a.fAngle );

	v[ 0 ] = a.vPos + r[ 0 ];
	v[ 1 ] = a.vPos + r[ 1 ];
	v[ 2 ] = a.vPos + r[ 2 ];
	v[ 3 ] = a.vPos + r[ 3 ];

	fMin = v[ 0 ] & axis;
	fMax = v[ 0 ] & axis;

	float	fTemp = v[ 1 ] & axis;

	if( fMin > fTemp )
	{
		fMin = fTemp;
	}
	else
	{
		if( fMax < fTemp )
		{
			fMax = fTemp;
		}
	}

	fTemp = v[ 2 ] & axis;

	if( fMin > fTemp )
	{
		fMin = fTemp;
	}
	else
	{
		if( fMax < fTemp )
		{
			fMax = fTemp;
		}
	}

	fTemp = v[ 3 ] & axis;

	if( fMin > fTemp )
	{
		fMin = fTemp;
	}
	else
	{
		if( fMax < fTemp )
		{
			fMax = fTemp;
		}
	}
}

bool clIntersect::funcIntersectAABBxAABB( clAABB & a0, clAABB & a1 )
{
	clVec2	vDel = a1.vPos - a0.vPos;	//    

	//    vDel   X  Y <=     
	return ( ( fabs( vDel.x ) <= ( a0.vSiz.x + a1.vSiz.x ) ) && 
			 ( fabs( vDel.y ) <= ( a0.vSiz.y + a1.vSiz.y ) ) );
}

bool clIntersect::funcIntersectSpherexSphere( clSphere & a0, clSphere & a1, clVec2 & v )
{
	float	fRadx2 = a0.fRadius + a1.fRadius;	//  
	v = a0.vPos - a1.vPos;		//   
	float	fDist = ( v ) & ( v );	//  

	//           
	if( fRadx2 * fRadx2 >= fDist )
	{
		float	f = sqrt( fDist );
		v /= f;		//  
		v *= ( fRadx2 - f );

		return true;
	}
	else
	{
		v.VecZero();

		return false;
	}
}

bool clIntersect::funcIntersectAABBxSphere( clAABB & a0, clSphere & a1, clVec2 & v )
{
	float	faMin, faMax;
	float	fbMin, fbMax;
	float	fD0, fD1;	//      
	float	fDepth, fTemp;
	clVec2	vDel = a0.vPos - a1.vPos;	//   

	//      X  Y
	clVec2	vAxis( 1.0f, 0.0f );
	funcGetIntervalAABB( a0, vAxis, faMin, faMax );
	funcGetIntervalSphere( a1, vAxis, fbMin, fbMax );
	if( ( faMin > fbMax ) || faMax < fbMin )
	{
		return false;
	}

	v = vAxis;
	fD0 = faMax - fbMin;
	fD1 = fbMax - faMin;
	fDepth = ( fD0 < fD1 ? fD0 : fD1 );

	vAxis.VecSet( 0.0f, 1.0f );
	funcGetIntervalAABB( a0, vAxis, faMin, faMax );
	funcGetIntervalSphere( a1, vAxis, fbMin, fbMax );
	if( ( faMin > fbMax ) || faMax < fbMin )
	{
		return false;
	}

	fD0 = faMax - fbMin;
	fD1 = fbMax - faMin;
	fTemp = ( fD0 < fD1 ? fD0 : fD1 );
	if( fDepth > fTemp )
	{
		fDepth = fTemp;
		v = vAxis;
	}

	//   4       
	// A B
	// C D
	//  B
	vAxis = a1.vPos - a0.vPos + a0.vSiz;
	vAxis.VecNormalize();
	funcGetIntervalAABB( a0, vAxis, faMin, faMax );
	funcGetIntervalSphere( a1, vAxis, fbMin, fbMax );
	if( ( faMin > fbMax ) || faMax < fbMin )
	{
		return false;
	}

	fD0 = faMax - fbMin;
	fD1 = fbMax - faMin;
	fTemp = ( fD0 < fD1 ? fD0 : fD1 );
	if( fDepth > fTemp )
	{
		fDepth = fTemp;
		v = vAxis;
	}

	//  C
	vAxis = a1.vPos - a0.vPos - a0.vSiz;
	vAxis.VecNormalize();
	funcGetIntervalAABB( a0, vAxis, faMin, faMax );
	funcGetIntervalSphere( a1, vAxis, fbMin, fbMax );
	if( ( faMin > fbMax ) || faMax < fbMin )
	{
		return false;
	}

	fD0 = faMax - fbMin;
	fD1 = fbMax - faMin;
	fTemp = ( fD0 < fD1 ? fD0 : fD1 );
	if( fDepth > fTemp )
	{
		fDepth = fTemp;
		v = vAxis;
	}

	//  A
	vAxis.VecSet( a1.vPos.x - a0.vPos.x - a0.vSiz.x, a1.vPos.y - a0.vPos.y + a0.vSiz.y );
	vAxis.VecNormalize();
	funcGetIntervalAABB( a0, vAxis, faMin, faMax );
	funcGetIntervalSphere( a1, vAxis, fbMin, fbMax );
	if( ( faMin > fbMax ) || faMax < fbMin )
	{
		return false;
	}

	fD0 = faMax - fbMin;
	fD1 = fbMax - faMin;
	fTemp = ( fD0 < fD1 ? fD0 : fD1 );
	if( fDepth > fTemp )
	{
		fDepth = fTemp;
		v = vAxis;
	}

	//  D
	vAxis.VecSet( a1.vPos.x - a0.vPos.x + a0.vSiz.x, a1.vPos.y - a0.vPos.y - a0.vSiz.y );
	vAxis.VecNormalize();
	funcGetIntervalAABB( a0, vAxis, faMin, faMax );
	funcGetIntervalSphere( a1, vAxis, fbMin, fbMax );
	if( ( faMin > fbMax ) || faMax < fbMin )
	{
		return false;
	}

	fD0 = faMax - fbMin;
	fD1 = fbMax - faMin;
	fTemp = ( fD0 < fD1 ? fD0 : fD1 );
	if( fDepth > fTemp )
	{
		fDepth = fTemp;
		v = vAxis;
	}

	// 
	if( ( v & vDel ) > 0.0f )
	{
		v *= - 1;
	}

	v *= fDepth;

	return true;
}

bool clIntersect::funcIntersectSpherexLine( clSphere & a0, clLine & a1, clVec2 & v )
{
	float	faMin, faMax;	//  
	float	fbMin, fbMax;	//  
	float	fD0, fD1;	//      
	float	fDepth, fTemp;
	clVec2	vDel = a0.vPos - a1.vMid;	//   

	//    A   B
	//      3 
	// 1     A
	clVec2	vAxis = ( a0.vPos - ( a1.vMid - a1.vDir * a1.fHalf ) );
	vAxis.VecNormalize();
	funcGetIntervalSphere( a0, vAxis, faMin, faMax );
	funcGetIntervalLine( a1, vAxis, fbMin, fbMax );
	if( ( faMin > fbMax ) || faMax < fbMin )
	{
		return false;
	}

	//   
	v = vAxis;
	fD0 = faMax - fbMin;
	fD1 = fbMax - faMin;
	fDepth = ( fD0 < fD1 ? fD0 : fD1 );

	// 2     B
	vAxis = ( a0.vPos - ( a1.vMid + a1.vDir * a1.fHalf ) );
	vAxis.VecNormalize();
	funcGetIntervalSphere( a0, vAxis, faMin, faMax );
	funcGetIntervalLine( a1, vAxis, fbMin, fbMax );
	if( ( faMin > fbMax ) || ( faMax < fbMin ) )
	{
		return false;
	}

	//   
	fD0 = faMax - fbMin;
	fD1 = fbMax - faMin;
	fTemp = ( fD0 < fD1 ? fD0 : fD1 );
	if( fDepth > fTemp )
	{
		fDepth = fTemp;
		v = vAxis;
	}

	// 3           
	vAxis = a1.vDir.VecNormal();
	vAxis.VecNormalize();
	funcGetIntervalSphere( a0, vAxis, faMin, faMax );
	funcGetIntervalLine( a1, vAxis, fbMin, fbMax );
	if( ( faMin > fbMax ) || ( faMax < fbMin ) )
	{
		return false;
	}

	//   
	fD0 = faMax - fbMin;
	fD1 = fbMax - faMin;
	fTemp = ( fD0 < fD1 ? fD0 : fD1 );
	if( fDepth > fTemp )
	{
		fDepth = fTemp;
		v = vAxis;
	}

	if( ( v & vDel ) > 0.0f )
	{
		v *= - 1;
	}

	v *= fDepth;

	//    ,   
	return true;
}

bool clIntersect::funcIntersectSpherexOBB( clSphere & a0, clOBB & a1, clVec2 & v )
{
	clSphere	sphere;
	clAABB		aabb;

	clVec2		vDel = a0.vPos - a1.vPos;	//    
	//     ,      OBB
	vDel.VecRotate( - a1.fAngle );
	sphere.vPos = a1.vPos + vDel;	//    
	sphere.fRadius = a0.fRadius;	// 
	//   AABB
	aabb.vPos = a1.vPos;
	aabb.vSiz = a1.vSiz;

	//    AABB x Sphere
	if( funcIntersectAABBxSphere( aabb, sphere, v ) )
	{
		v.VecRotate( a1.fAngle );

		return true;
	}
	else
	{
		return false;
	}
}

bool clIntersect::funcIntersectAABBxLine( clAABB & a0, clLine & a1, clVec2 & v )
{
	float	faMin, faMax;	//  AABB
	float	fbMin, fbMax;	//  
	float	fD0, fD1;	//      
	float	fDepth, fTemp;
	clVec2	vDel = a0.vPos - a1.vMid;	//   

	//  AABB      
	// fbMin -          
	clVec2	vAxis = a1.vDir.VecNormal();
	funcGetIntervalAABB( a0, vAxis, faMin, faMax );
	fbMin = a1.vMid & ( vAxis );
	if( ( fbMin < faMin ) || ( fbMin > faMax ) )
	{
		return false;
	}

	v = vAxis;
	fD0 = faMax - fbMin;
	fD1 = fbMin - faMin;
	fDepth = ( fD0 < fD1 ? fD0 : fD1 );

	//    X
	vAxis.VecSet( 1.0f, 0.0f );
	funcGetIntervalAABB( a0, vAxis, faMin, faMax );
	funcGetIntervalLine( a1, vAxis, fbMin, fbMax );
	if( ( fbMax < faMin ) || ( fbMin > faMax ) )
	{
		return false;
	}

	//   
	fD0 = faMax - fbMin;
	fD1 = fbMax - faMin;
	fTemp = ( fD0 < fD1 ? fD0 : fD1 );
	if( fDepth > fTemp )
	{
		fDepth = fTemp;
		v = vAxis;
	}

	//    Y
	vAxis.VecSet( 0.0f, 1.0f );
	funcGetIntervalAABB( a0, vAxis, faMin, faMax );
	funcGetIntervalLine( a1, vAxis, fbMin, fbMax );
	if( ( fbMax < faMin ) || ( fbMin > faMax ) )
	{
		return false;
	}

	//   
	fD0 = faMax - fbMin;
	fD1 = fbMax - faMin;
	fTemp = ( fD0 < fD1 ? fD0 : fD1 );
	if( fDepth > fTemp )
	{
		fDepth = fTemp;
		v = vAxis;
	}

	//         
	if( ( vDel & v ) > 0.0f ) 
        v *= - 1;

	v *= fDepth;

	return true;
}

bool clIntersect::funcIntersectOBBxLine( clOBB & a0, clLine & a1, clVec2 & v )
{
	clAABB		aabb;
	clLine		line;

	aabb.vPos = a0.vPos;
	aabb.vSiz = a0.vSiz;

	//       OBB
	clVec2		vDel = a1.vMid - a0.vPos;
	vDel.VecRotate( - a0.fAngle );
	line.vMid = a0.vPos + vDel;
	
	//  
	vDel = a1.vDir;
	vDel.VecRotate( - a0.fAngle );
	line.vDir = vDel;

	//    
	line.fHalf = a1.fHalf;

	if( funcIntersectAABBxLine( aabb, line, v ) )
	{
		//       
		v.VecRotate( a0.fAngle );

		return true;
	}
	else
	{
		return false;
	}
}

bool clIntersect::funcIntersectLinexLine( clLine & a0, clLine & a1, clVec2 & v )
{
	float	fMin, fMax;
	float	fD0, fD1;	//      
	float	fDepth, fTemp;
	clVec2	vDel = a0.vMid - a1.vMid;	//   

	//     a1
	clVec2	vAxis = a1.vDir.VecNormal();
	vAxis.VecNormalize();
	funcGetIntervalLine( a0, vAxis, fMin, fMax );
	float	fDot = a1.vMid & vAxis;
	if( ( fDot < fMin ) || ( fDot > fMax ) )
	{
		return false;
	}

	// 
	v = vAxis;
	fD0 = fDot - fMin;
	fD1 = fMax - fDot;
	fDepth = ( fD0 < fD1 ? fD0 : fD1 );

	//     a0
	vAxis = a0.vDir.VecNormal();
	vAxis.VecNormalize();
	funcGetIntervalLine( a1, vAxis, fMin, fMax );
	fDot = a0.vMid & vAxis;
	if( ( fDot < fMin ) || ( fDot > fMax ) )
	{
		return false;
	}

	// 
	fD0 = fDot - fMin;
	fD1 = fMax - fDot;
	fTemp = ( fD0 < fD1 ? fD0 : fD1 );
	if( fDepth > fTemp )
	{
		fDepth = fTemp;
		v = vAxis;
	}

	//  
	if( ( vDel & v ) > 0.0f ) 
        v *= - 1;

	v *= fDepth;

	return true;
}

bool clIntersect::funcIntersectAABBxOBB( clAABB & a0, clOBB & a1, clVec2 & v )
{
	float	faMin, faMax;	//  AABB
	float	fbMin, fbMax;	//  
	float	fD0, fD1;	//      
	float	fDepth, fTemp;
	clVec2	vDel = a0.vPos - a1.vPos;	//   

	//     AABB,    1, 0  0, 1      
	clVec2	vAxis( 1.0f, 0.0f );
	funcGetIntervalAABB( a0, vAxis, faMin, faMax );
	funcGetIntervalOBB( a1, vAxis, fbMin, fbMax );

	if( ( faMax < fbMin ) || ( faMin > fbMax ) )
	{
		return false;
	}

	// 
	v = vAxis;
	fD0 = faMax - fbMin;
	fD1 = fbMax - faMin;
	fDepth = ( fD0 < fD1 ? fD0 : fD1 );

	vAxis.VecSet( 0.0f, 1.0f );
	funcGetIntervalAABB( a0, vAxis, faMin, faMax );
	funcGetIntervalOBB( a1, vAxis, fbMin, fbMax );

	if( ( faMax < fbMin ) || ( faMin > fbMax ) )
	{
		return false;
	}

	// 
	fD0 = faMax - fbMin;
	fD1 = fbMax - faMin;
	fTemp = ( fD0 < fD1 ? fD0 : fD1 );
	if( fDepth > fTemp )
	{
		fDepth = fTemp;
		v = vAxis;
	}

	//    OBB

	float		fCos = cos( a1.fAngle );
	float		fSin = sin( a1.fAngle );

	vAxis.VecSet( fCos, fSin );
	funcGetIntervalAABB( a0, vAxis, faMin, faMax );
	funcGetIntervalOBB( a1, vAxis, fbMin, fbMax );

	if( ( faMax < fbMin ) || ( faMin > fbMax ) )
	{
		return false;
	}

	// 
	fD0 = faMax - fbMin;
	fD1 = fbMax - faMin;
	fTemp = ( fD0 < fD1 ? fD0 : fD1 );
	if( fDepth > fTemp )
	{
		fDepth = fTemp;
		v = vAxis;
	}

	vAxis.VecSet( - fSin, fCos );
	funcGetIntervalAABB( a0, vAxis, faMin, faMax );
	funcGetIntervalOBB( a1, vAxis, fbMin, fbMax );

	if( ( faMax < fbMin ) || ( faMin > fbMax ) )
	{
		return false;
	}

	// 
	fD0 = faMax - fbMin;
	fD1 = fbMax - faMin;
	fTemp = ( fD0 < fD1 ? fD0 : fD1 );
	if( fDepth > fTemp )
	{
		fDepth = fTemp;
		v = vAxis;
	}

	// It's done
	if( ( vDel & v ) > 0.0f ) 
        v *= - 1;

	v *= fDepth;

	return true;
}

bool clIntersect::funcIntersectOBBxOBB( clOBB & a0, clOBB & a1, clVec2 & v )
{
	clVec2	vDel = a1.vPos - a0.vPos;

	// AABB   OBB
	clAABB	aabb;
	aabb.vPos = a0.vPos;
	aabb.vSiz = a0.vSiz;

	//   OBB    
	clOBB	obb;
	vDel.VecRotate( - a0.fAngle );
	obb.fAngle = a1.fAngle - a0.fAngle;
	obb.vPos = a0.vPos + vDel;
	obb.vSiz = a1.vSiz;

	if( funcIntersectAABBxOBB( aabb, obb, v ) )
	{
		v.VecRotate( a0.fAngle );

		return true;
	}
	else
	{
		return false;
	}
}